#!/bin/sh -e
# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright 2023 Oliver Smith
SCRIPT_DIR="$(dirname "$(realpath "$0")")"
TOP_DIR="$(realpath "$SCRIPT_DIR/..")"
WORK_DIR="$(pmbootstrap -q config work)"
ROOTFS_DIR="$WORK_DIR/chroot_native/home/pmos/rootfs"
PMAPORTS_DIR="$(pmbootstrap -q config aports)"
DEVICE="$(pmbootstrap -q config device)"

ONDEV2="$(realpath "$SCRIPT_DIR"/../ondev2)"
TEMP_DIR="$TOP_DIR/_temp"
OS_IMAGE="$TEMP_DIR/os_image.img"
STORAGE_INSTALLER_DEV="/dev/vda"
STORAGE_SECOND_DEV="/dev/vdb"

ARG_ALL_CASES=0
ARG_BUILD_BOOT_DEPLOY=0
ARG_CASES=""
ARG_CASES_DEFAULT="01-sd-to-sd-simple"
ARG_INTERACTIVE=0
ARG_BUILD_PMOS_MKINITFS=0
ARG_RUN_INITFS_SHELL=0
ARG_SKIP_BUILD=0

mkdir -p "$TEMP_DIR"

if ! [ -e "$OS_IMAGE" ]; then
	echo "ERROR: missing OS image: $OS_IMAGE"
	echo "Either put one there or create a pmOS image with pmbootstrap by"
	echo "running distro-specific/postmarketOS/create-test-image.sh"
	exit 1
fi

# Source deviceinfo for deviceinfo_kernel_cmdline
. "$PMAPORTS_DIR"/device/*/device-"$DEVICE"/deviceinfo

parse_args() {
	while getopts 'abc:imrsh' OPTION; do
		case "$OPTION" in
		a)
			echo "Running all cases"
			ARG_ALL_CASES=1
			;;
		b)
			echo "Building boot-deploy"
			ARG_BUILD_BOOT_DEPLOY=1
			;;
		c)
			echo "Running case: $OPTARG"
			ARG_CASES="$ARG_CASES $OPTARG"
			;;
		i)
			echo "Loading one case config and running interactively"
			ARG_INTERACTIVE=1
			;;
		m)
			echo "Building postmarketos-mkinitfs"
			ARG_BUILD_PMOS_MKINITFS=1
			;;
		r)
			echo "Running shell inside initramfs hook"
			ARG_RUN_INITFS_SHELL=1
			;;
		s)
			echo "Skipping package and install image building"
			ARG_SKIP_BUILD=1
			;;
		h|*)
			echo "usage: $(basename $0) [-a] [[-c CASE] [-c CASE2 ... ]]" \
				"[-i] [-m] [-r] [-s] [-h]"
			echo "arguments:"
			echo "  -a       run all tests"
			echo "  -b       build boot-deploy from source"
			echo "  -c CASE  run this specific case (default: $ARG_CASES_DEFAULT)"
			echo "  -i       load case config only and go interactive"
			echo "  -m       build postmarketos-mkinitfs from source"
			echo "  -r       run shell inside initramfs hook"
			echo "  -s       skip package and install image build"
			echo "           (case and -r same as last run)"
			echo "  -h       show help"
			exit 1
			;;
		esac
	done

	if [ "$ARG_ALL_CASES" = 1 ]; then
		if [ -n "$ARG_CASES" ]; then
			echo "ERROR: can't combine -a with -c"
			exit 1
		fi

		for i in test/cases/*.cfg; do
			ARG_CASES="$ARG_CASES $(basename "$i" | cut -d. -f 1)"
		done
	fi

	if [ -z "$ARG_CASES" ]; then
		ARG_CASES="$ARG_CASES_DEFAULT"
	fi
}

clean_up() {
	if [ "$ARG_SKIP_BUILD" = 1 ]; then
		return
	fi

	pmbootstrap -q -y zap -p
}

move_install_chroot() {
	# Move the installer chroot (generated by "pmbootstrap install") to
	# the regular device chroot path, so "pmbootstrap qemu" will use the
	# kernel and initramfs from there.

	pmbootstrap -q shutdown

	! test -d "$WORK_DIR"/chroot_rootfs_"$DEVICE"

	sudo mv "$WORK_DIR"/chroot_installer_"$DEVICE" \
		"$WORK_DIR"/chroot_rootfs_"$DEVICE"

	sed -i "s/installer_$DEVICE/rootfs_$DEVICE/g" "$WORK_DIR"/workdir.cfg
}

build_install_image() {
	local test_dir="$TEMP_DIR/test_dir"
	local test_files="$TEMP_DIR/ondev2-test.files"

	if [ "$ARG_SKIP_BUILD" = 1 ]; then
		return
	fi

	# First time building the install image in this test session
	if ! [ -e "$WORK_DIR"/chroot_native/home/pmos/rootfs/*.img ]; then
		# Remove big files from test/run-test-host-losetup.sh so they
		# don't get copied into the chroot when building the package
		# from source
		rm -f "$TEMP_DIR"/storage_installer.img \
			"$TEMP_DIR"/storage_second.img

		pmbootstrap -q build --src="$TOP_DIR" postmarketos-ondev

		if [ "$ARG_BUILD_PMOS_MKINITFS" = 1 ]; then
			pmbootstrap -q build \
				--src="$TOP_DIR/../postmarketos-mkinitfs" \
				postmarketos-mkinitfs
		fi

		if [ "$ARG_BUILD_BOOT_DEPLOY" = 1 ]; then
			pmbootstrap -q build \
				--src="$TOP_DIR/../boot-deploy" \
				boot-deploy
		fi
	fi

	# Put tests into a tarball
	mkdir -p "$test_dir"/test
	cp -r "$SCRIPT_DIR"/common.exp "$SCRIPT_DIR"/cases "$test_dir"/test
	tar -C "$test_dir" -cvzf "$test_dir/tests.tar.gz" test

	# Add test related files to the initfs
	( echo "/usr/bin/expect"
	  echo "/usr/lib/tcl*"
	  echo "/opt/run-test.sh"
	  echo "/opt/tests.tar.gz" ) > "$TEMP_DIR/ondev2-test.files"

	pmbootstrap -q install \
		--ondev \
		--ondev2-image "$OS_IMAGE" \
		--cp \
			"$TEMP_DIR/cmdline.txt:/etc/boot/cmdline.txt" \
			"$TEMP_DIR/ondev2.test.cfg:/etc/ondev2.test.cfg" \
			"$test_files:/usr/share/mkinitfs/files-extra/ondev2-test.files" \
			"$SCRIPT_DIR/lib/run-test-from-kernel-cmdline.sh:/opt/run-test.sh" \
			"$test_dir/tests.tar.gz:/opt/tests.tar.gz"

	move_install_chroot
}

fail_test() {
	set +x
	echo "---"
	echo "TEST FAILED!"
	echo
	set -x

	exit 1
}

run_test() {
	local i
	local test_exp="$SCRIPT_DIR/cases/$1.exp"
	local test_cfg="$SCRIPT_DIR/cases/$1.cfg"
	local log="$TEMP_DIR/log"
	local cmdline=""

	for i in $test_exp $test_cfg; do
		if ! [ -e "$i" ]; then
			set +x
			echo "ERROR: file not found: $i"
			exit 1
		fi
	done

	# Prepare ondev2.test.cfg. The ondev-prepare.sh script generates a
	# ondev2.cfg based on the qemu device's deviceinfo file and then
	# appends the ondev2.test.cfg file to apply overrides from the test
	# case for storage devices etc.
	sed \
		-e "s#@TEST_STORAGE_INSTALLER@#$STORAGE_INSTALLER_DEV#g" \
		-e "s#@TEST_STORAGE_SECOND@#$STORAGE_SECOND_DEV#g" \
		"$test_cfg" > "$TEMP_DIR/ondev2.test.cfg"

	if [ "$ARG_INTERACTIVE" = 0 ]; then
		cmdline="$cmdline ONDEV2_TEST=$(basename "$test_exp")"
	fi

	if [ "$ARG_RUN_INITFS_SHELL" = 1 ]; then
		cmdline="$cmdline ONDEV2_RUN_INITFS_SHELL"
	fi

	# boot-deploy will append this to the cmdline in grub.cfg
	echo "$cmdline" > "$TEMP_DIR/cmdline.txt"

	build_install_image

	# Run pmbootstrap command outside of expect right before starting
	# pmbootstrap qemu, in case there's a sudo prompt
	pmbootstrap -q chroot -- true

	export DISPLAY=${DISPLAY:-:0}

	rm -f "$log"
	pmbootstrap -q qemu --second-storage --efi | tee "$log"

	if grep -q "ERROR" "$log"; then
		test_failed
	fi

	if ! grep -q "You have successfully installed" "$log"; then
		test_failed
	fi

	if grep -q ":: Remove SD" "$log"; then
		# Emulate removing SD card
		sudo mv "$ROOTFS_DIR/$DEVICE-2nd.img" "$ROOTFS_DIR/$DEVICE.img"
		# Boot the finished installation
		pmbootstrap -q qemu --efi | tee -a "$log"

		if grep -q "ERROR" "$log"; then
			test_failed
		fi
	fi

	set +x
	echo "---"
	echo "TEST SUCCESSFUL!"
	echo
	set -x
}

parse_args "$@"

set -x

clean_up

for i in $ARG_CASES; do
	run_test "$i"
done
